open Syntax
open Value

(* 実際の計算をする関数 *)
(* Refunctionalized interpreter : eval6 *)

(* 初期継続 *)
let idc v s t m = match s with
    SNil ->
    begin match t with
        TNil ->
        begin match m with
            MNil -> v
          | MCons ((c, s, t) ,m) -> c v s t m
        end
      | Trail (c, s) -> c v s TNil m
    end
  | _ -> failwith "stack error"

(* cons : c -> s -> t -> t *)
let rec cons c s t = match t with
    TNil -> Trail (c, s)
  | Trail (c', s') ->
    Trail (
      (fun v s t m ->
         begin match s with
             SAppend (s, s') -> c v s (cons c' s' t) m
           | _ -> failwith "stack error"
         end),
      SAppend (s, s'))

(* apnd : t -> t -> t *)
let apnd t0 t1 = match t0 with
    TNil -> t1
  | Trail (c, s) -> cons c s t1

(* f6 : e -> string list -> v list -> s -> c -> t -> m -> v *)
let rec f6 e xs vs s c t m = match e with
    Num (n) -> c (VNum (n)) s t m 
  | Var (x) -> c (List.nth vs (Env.offset x xs)) s t m
  | Op (e0, op, e1) ->
    f6 e0 xs vs (SCons (VEnv (vs), s)) (fun v0 s0 t0 m0 ->
        begin match s0 with
            SCons (VEnv (vs), s0) ->
            f6 e1 xs vs (SCons (v0, s0)) (fun v1 s1 t1 m1 ->
                begin match s1 with
                    SCons (v0, s1) ->
                    begin match (v0, v1) with
                        (VNum (n0), VNum (n1)) ->
                        begin match op with
                            Plus -> c (VNum (n0 + n1)) s1 t1 m1
                          | Minus -> c (VNum (n0 - n1)) s1 t1 m1
                          | Times -> c (VNum (n0 * n1)) s1 t1 m1
                          | Divide ->
                            if n1 = 0 then failwith "Division by zero"
                            else c (VNum (n0 / n1)) s1 t1 m1
                        end
                      | _ -> failwith "op error"
                    end
                  | _ -> failwith "stack error"
                end) t0 m0
          | _ -> failwith "stack error"
        end) t m
  | Fun (x, e) ->
    let vfun = VFun (fun v s' c' t' m' -> f6 e (x :: xs) (v :: vs) s' c' t' m')
    in c vfun s t m
  | App (e0, e1) ->
    f6 e0 xs vs (SCons (VEnv (vs), s)) (fun v0 s0 t0 m0 ->
        begin match s0 with
            SCons (VEnv (vs), s0) ->
            f6 e1 xs vs (SCons (v0, s0)) (fun v1 s1 t1 m1 ->
                begin match s1 with
                    SCons (v0, s1) ->
                    begin match v0 with
                        VFun (f) -> f v1 s1 c t1 m1
                      | VContS (c', s', t') ->
                        c' v1 s' t' (MCons ((c, s1, t1), m1))
                      | VContC (c', s', t') ->
                        c' v1 s' (apnd t' (cons c s1 t1)) m1
                      | _ -> failwith "not app"
                    end
                  | _ -> failwith "stack error"
                end) t0 m0
          | _ -> failwith "stack error"
        end) t m
  | Shift (x, e) -> f6 e (x :: xs) (VContS (c, s, t) :: vs) SNil idc TNil m
  | Control (x, e) -> f6 e (x :: xs) (VContC (c, s, t) :: vs) SNil idc TNil m
  | Shift0 (x, e) ->
    begin match m with
        MCons ((c0, s0, t0), m0) ->
        f6 e (x :: xs) (VContS (c, s, t) :: vs) s0 c0 t0 m0
      | _ -> failwith "meta cont error"
    end
  | Control0 (x, e) ->
    begin match m with
        MCons ((c0, s0, t0), m0) ->
        f6 e (x :: xs) (VContC (c, s, t) :: vs) s0 c0 t0 m0
      | _ -> failwith "meta cont error"
    end
  | Reset (e) -> f6 e xs vs SNil idc TNil (MCons ((c, s, t), m))

(* f : e -> v *)
let f expr = f6 expr [] [] SNil idc TNil MNil
